#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <libxml/parser.h>

#include "specimen.h"
#include "patch.h"

int beef_write(char *name)
{
     char tmp[256];
     int val;
     int *id;
     int i, count;
     xmlDocPtr doc;
     xmlNodePtr rootnode;
     xmlNodePtr xmlpatch;

     debug("Writing current bank to file %s\n", name);

     /* create the guts of the doc */
     doc = xmlNewDoc("1.0");
     rootnode = xmlNewDocNode(doc, NULL, (const xmlChar *)"beef", NULL);
     xmlDocSetRootElement(doc, rootnode);

     /* walk the walk */
     count = patch_dump(&id);
     for (i = 0; i < count; i++) {
	  xmlpatch = xmlNewTextChild(rootnode, NULL, (const xmlChar *)"patch", NULL);
	  
	  xmlNewTextChild(xmlpatch, NULL, (const xmlChar *)"name", (const xmlChar *)patch_name_get(id[i]));

	  xmlNewTextChild(xmlpatch, NULL, (const xmlChar *)"file",
			       (const xmlChar *)patch_sample_name_get(id[i]));
	  
	  sprintf(tmp, "%d", patch_channel_get(id[i]));
	  xmlNewTextChild(xmlpatch, NULL, (const xmlChar *)"channel", (const xmlChar *)tmp);

	  sprintf(tmp, "%d", patch_note_get(id[i]));
	  xmlNewTextChild(xmlpatch, NULL, (const xmlChar *)"note", (const xmlChar *)tmp);
	  
	  sprintf(tmp, "%d", patch_volume_get(id[i]));
	  xmlNewTextChild(xmlpatch, NULL, (const xmlChar *)"volume", (const xmlChar *)tmp);
	  
	  sprintf(tmp, "%d", patch_pan_get(id[i]));
	  xmlNewTextChild(xmlpatch, NULL, (const xmlChar *)"pan", (const xmlChar *)tmp);

	  sprintf(tmp, "%d", patch_play_mode_get(id[i]));
	  xmlNewTextChild(xmlpatch, NULL, (const xmlChar *)"play_mode", (const xmlChar *)tmp);

	  sprintf(tmp, "%d", patch_cut_get(id[i]));
	  xmlNewTextChild(xmlpatch, NULL, (const xmlChar *)"cut", (const xmlChar *)tmp);

	  sprintf(tmp, "%d", patch_cut_by_get(id[i]));
	  xmlNewTextChild(xmlpatch, NULL, (const xmlChar *)"cut_by", (const xmlChar *)tmp);

	  sprintf(tmp, "%d", patch_range_get(id[i]));
	  xmlNewTextChild(xmlpatch, NULL, (const xmlChar *)"range", (const xmlChar *)tmp);

	  sprintf(tmp, "%d", patch_lower_note_get(id[i]));
	  xmlNewTextChild(xmlpatch, NULL, (const xmlChar *)"lower_note", (const xmlChar *)tmp);

	  sprintf(tmp, "%d", patch_upper_note_get(id[i]));
	  xmlNewTextChild(xmlpatch, NULL, (const xmlChar *)"upper_note", (const xmlChar *)tmp);

	  sprintf(tmp, "%d", patch_sample_start_get(id[i]));
	  xmlNewTextChild(xmlpatch, NULL, (const xmlChar *)"sample_start", (const xmlChar *)tmp);

	  sprintf(tmp, "%d", patch_sample_stop_get(id[i]));
	  xmlNewTextChild(xmlpatch, NULL, (const xmlChar *)"sample_stop", (const xmlChar *)tmp);

	  sprintf(tmp, "%d", patch_loop_start_get(id[i]));
	  xmlNewTextChild(xmlpatch, NULL, (const xmlChar *)"loop_start", (const xmlChar *)tmp);

	  sprintf(tmp, "%d", patch_loop_stop_get(id[i]));
	  xmlNewTextChild(xmlpatch, NULL, (const xmlChar *)"loop_stop", (const xmlChar *)tmp);

	  sprintf(tmp, "%f", patch_filter_get(id[i]));
	  xmlNewTextChild(xmlpatch, NULL, (const xmlChar *)"filter", (const xmlChar *)tmp);

	  sprintf(tmp, "%f", patch_resonance_get(id[i]));
	  xmlNewTextChild(xmlpatch, NULL, (const xmlChar *)"resonance", (const xmlChar *)tmp);

	  sprintf(tmp, "%f", patch_attack_get(id[i]));
	  xmlNewTextChild(xmlpatch, NULL, (const xmlChar *)"attack", (const xmlChar *)tmp);

	  sprintf(tmp, "%f", patch_decay_get(id[i]));
	  xmlNewTextChild(xmlpatch, NULL, (const xmlChar *)"decay", (const xmlChar *)tmp);

	  sprintf(tmp, "%f", patch_sustain_get(id[i]));
	  xmlNewTextChild(xmlpatch, NULL, (const xmlChar *)"sustain", (const xmlChar *)tmp);

	  sprintf(tmp, "%f", patch_release_get(id[i]));
	  xmlNewTextChild(xmlpatch, NULL, (const xmlChar *)"release", (const xmlChar *)tmp);
     }
     
     /* write the file to disk */
     val = xmlSaveFormatFile(name, doc, 1);
     xmlFreeDoc(doc);

     return (val)? val: 0;
}

int beef_read(char *path)
{
     xmlDocPtr doc;
     xmlNodePtr rootnode;
     xmlNodePtr xmlpatch;
     xmlNodePtr cur;
     xmlChar *key;
     char *name, *file;
     int channel, note, id, volume, pan, cut, cut_by, range, lower_note, upper_note, 
	  sample_start, sample_stop, loop_start, loop_stop;
     float filter, reso, a, d, s, r;
     play_mode_t mode;

     debug("Loading bank from file %s\n", path);

     doc = xmlParseFile(path);
     if (doc == NULL) {
	  errmsg("Failed to parse %s\n", path);
	  return -1;
     }

     rootnode = xmlDocGetRootElement(doc);
     if (rootnode == NULL) {
	  errmsg("%s is empty\n", path);
	  xmlFreeDoc(doc);
	  return -1;
     }

     if (xmlStrcmp(rootnode->name, (const xmlChar *)"beef") != 0) {
	  errmsg("%s is not a valid 'beef' file\n", path);
	  xmlFreeDoc(doc);
	  return -1;
     }


     patch_destroy_all();
     for (xmlpatch = rootnode->xmlChildrenNode; xmlpatch != NULL;
	  xmlpatch = xmlpatch->next) {

	  name = file = key = NULL;
	  channel = note = id = volume = pan = cut = cut_by = range = lower_note = upper_note = \
	       sample_start = sample_stop = loop_start = loop_stop = filter = reso = \
	       a = d = s = r = -1;
	  mode = 0;
	  if (!xmlStrcmp(xmlpatch->name, (const xmlChar *)"patch")) {
	       for (cur = xmlpatch->xmlChildrenNode; cur != NULL; cur = cur->next) {
		    if (!xmlStrcmp(cur->name, (const xmlChar *)"name")) {
			 key = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
			 name = strdup((const char *)key);
		    
		    } else if (!xmlStrcmp(cur->name, (const xmlChar *)"file")) {
			 key = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
			 file = strdup((const char *)key);
		    
		    } else if (!xmlStrcmp(cur->name, (const xmlChar *)"channel")) {
			 key = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
			 channel = atoi((const char *)key);
		    
		    } else if (!xmlStrcmp(cur->name, (const xmlChar *)"note")) {
			 key = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
			 note = atoi((const char *)key);
		    
		    } else if (!xmlStrcmp(cur->name, (const xmlChar *)"volume")) {
			 key = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
			 volume = atoi((const char *)key);
		    
		    } else if (!xmlStrcmp(cur->name, (const xmlChar *)"pan")) {
			 key = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
			 pan = atoi((const char*)key);
		    
		    } else if (!xmlStrcmp(cur->name, (const xmlChar *)"play_mode")) {
			 key = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
			 mode = atoi((const char*)key);
		    
		    } else if (!xmlStrcmp(cur->name, (const xmlChar *)"cut")) {
			 key = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
			 cut = atoi((const char*)key);
		    
		    } else if (!xmlStrcmp(cur->name, (const xmlChar *)"cut_by")) {
			 key = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
			 cut_by = atoi((const char*)key);
		    
		    } else if (!xmlStrcmp(cur->name, (const xmlChar *)"range")) {
			 key = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
			 range = atoi((const char*)key);
		    
		    } else if (!xmlStrcmp(cur->name, (const xmlChar *)"lower_note")) {
			 key = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
			 lower_note = atoi((const char*)key);
		    
		    } else if (!xmlStrcmp(cur->name, (const xmlChar *)"upper_note")) {
			 key = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
			 upper_note = atoi((const char*)key);
		    
		    } else if (!xmlStrcmp(cur->name, (const xmlChar *)"sample_start")) {
			 key = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
			 sample_start = atoi((const char*)key);
		    
		    } else if (!xmlStrcmp(cur->name, (const xmlChar *)"sample_stop")) {
			 key = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
			 sample_stop = atoi((const char*)key);
		    
		    } else if (!xmlStrcmp(cur->name, (const xmlChar *)"loop_start")) {
			 key = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
			 loop_start = atoi((const char*)key);
		    
		    } else if (!xmlStrcmp(cur->name, (const xmlChar *)"loop_stop")) {
			 key = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
			 loop_stop = atoi((const char*)key);
		    
		    } else if (!xmlStrcmp(cur->name, (const xmlChar *)"filter")) {
			 key = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
			 filter = atof((const char*)key);
		    
		    } else if (!xmlStrcmp(cur->name, (const xmlChar *)"resonance")) {
			 key = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
			 reso = atof((const char*)key);
		    
		    } else if (!xmlStrcmp(cur->name, (const xmlChar *)"attack")) {
			 key = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
			 a = atof((const char*)key);

		    } else if (!xmlStrcmp(cur->name, (const xmlChar *)"decay")) {
			 key = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
			 d = atof((const char*)key);
		    
		    } else if (!xmlStrcmp(cur->name, (const xmlChar *)"sustain")) {
			 key = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
			 s = atof((const char*)key);
		    
		    } else if (!xmlStrcmp(cur->name, (const xmlChar *)"release")) {
			 key = xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
			 r = atof((const char*)key);
		    }
		    if (key != NULL) {
			 xmlFree(key);
			 key = NULL;
		    }
	       }

	       if (name != NULL) { /* must have at least a name to be valid */
		    id = patch_create(name);
		    free(name);
		    name = NULL; /* this is paranoid... but so am I */
		    if ((file != NULL) && strlen(file)) {
			 patch_sample_load(id, file);
			 free(file);
			 file = NULL;
		    }
		    if (channel >= 0)
			 patch_channel_set(id, channel);

		    if (note >= 0)
			 patch_note_set(id, note);

		    if (volume >= 0)
			 patch_volume_set(id, volume);

		    if (pan >= 0)
			 patch_pan_set(id, pan);

		    if (mode > 0)
			 patch_play_mode_set(id, mode);

		    if (cut > 0)
			 patch_cut_set(id, cut);

		    if (cut_by > 0)
			 patch_cut_by_set(id, cut_by);

		    if (range > 0)
			 patch_range_set(id, range);

		    if (lower_note >= 0)
			 patch_lower_note_set(id, lower_note);

		    if (upper_note >= 0)
			 patch_upper_note_set(id, upper_note);

		    if (sample_start >= 0)
			 patch_sample_start_set(id, sample_start);

		    if (sample_stop >= 0)
			 patch_sample_stop_set(id, sample_stop);

		    if (loop_start >= 0)
			 patch_loop_start_set(id, loop_start);

		    if (loop_stop >= 0)
			 patch_loop_stop_set(id, loop_stop);

		    if (filter >= 0)
			 patch_filter_set(id, filter);

		    if (reso >= 0)
			 patch_resonance_set(id, reso);

		    if (a >= 0)
			 patch_attack_set(id, a);

		    if (d >= 0)
			 patch_decay_set(id, d);
		    
		    if (s >= 0)
			 patch_sustain_set(id, s);

		    if (r >= 0)
			 patch_release_set(id, r);
	       }
	  }
     }

     xmlFreeDoc(doc);
     return 0;
}

